#include "loader.hpp"

#include <cmath>
#include <cstdio>
#include <fstream>
#include <string>
#include <iostream>
#include <vector>

#include <opencv2/core.hpp>
#include <opencv2/imgcodecs.hpp>

#include <experimental/filesystem>
namespace fs = std::experimental::filesystem;

cv::Mat read_pfm(const std::string &filename) {
	cv::Mat im;
	FILE * fp;

	fp = fopen(filename.c_str(), "rb");
	char buf[32];
	int fsize;
	int width;
	int height;
	float scale;

	if (fp == NULL) { return im; }

	if (fscanf(fp, "%31s", buf) == 0 || strcmp(buf, "Pf")) { goto cleanup; }

	if (fscanf(fp, "%31s", buf) == 0) { goto cleanup; }
	width = atoi(buf);

	if (fscanf(fp, "%31s", buf) == 0) { goto cleanup; }
	height = atoi(buf);

	if (fscanf(fp, " %31s", buf) == 0) { goto cleanup; }
	scale = atof(buf);

	im.create(height, width, CV_32FC1);

	fseek(fp, 0, SEEK_END);
	fseek(fp, ftell(fp)-width*height*sizeof(float), SEEK_SET);

	for (int y = 0; y < height; y++) {
		float* im_ptr = im.ptr<float>(height-y-1);
		int nread = 0;
		do {
			nread += fread(im_ptr+nread, sizeof(float), width-nread, fp);
			if (ferror(fp)) { goto cleanup; }
		}
		while (nread != width);
	}

	cleanup:
	fclose(fp);
	return im;
}

static void parse_camera_parameters(const std::string &v, MiddEvalCalib &calib) {
	std::string mat = std::string(v.substr(1, v.size()-2));
	std::string row;

	std::vector<float> values;

	for (int _ = 0; _ < 3; _++) {
		size_t pos = mat.find(";");
		row = mat.substr(0, pos);

		std::istringstream sstr(row);
		for(std::string val; sstr >> val;) {
			values.push_back(atof(val.c_str()));
		}

		mat.erase(0, pos+1);
	}

	calib.f = values[0];
	calib.cx = values[2];
	calib.cy = values[5];
}

MiddEvalCalib read_calibration(const std::string &filename) {
	MiddEvalCalib calib;
	memset(&calib, 0, sizeof(calib));

	std::ifstream f(filename);

	for(std::string line; std::getline(f, line);) {
		auto m = line.find("=");
		if (m == std::string::npos) { continue; }
		auto k = line.substr(0, m);
		auto v = line.substr(m+1);

		if (k == "baseline") {
			calib.baseline = atof(v.c_str());
		}
		else if (k == "doffs") {
			calib.doffs = atof(v.c_str());
		}
		else if (k == "ndisp") {
			calib.ndisp = atoi(v.c_str());
		}
		else if (k == "vmin") {
			calib.vmin = atoi(v.c_str());
		}
		else if (k == "vmax") {
			calib.vmax = atoi(v.c_str());
		}
		else if (k == "width") {
			calib.width = atoi(v.c_str());
		}
		else if (k == "height") {
			calib.height = atoi(v.c_str());
		}
		else if (k == "cam0") {
			parse_camera_parameters(v, calib);
		}
	}

	if (calib.vmax == 0) {
		calib.vmax = calib.ndisp;
	}

	return calib;
}

MiddleburyData load_input(const fs::path &path) {
	cv::Mat imL;
	cv::Mat imR;
	cv::Mat gtL;
	cv::Mat maskL;

	imL = cv::imread(path/"im0.png", cv::IMREAD_COLOR);
	imR = cv::imread(path/"im1.png", cv::IMREAD_COLOR);
	gtL = read_pfm(path/"disp0.pfm");
	if (gtL.empty()) {
		gtL = read_pfm(path/std::string("disp0GT.pfm"));
	}
	if (gtL.empty()) {
		gtL.create(imL.size(), CV_32FC1);
		gtL.setTo(cv::Scalar(0.0f));
	}

	maskL = cv::imread(path/std::string("mask0nocc.png"), cv::IMREAD_GRAYSCALE);
	if (maskL.empty()) {
		maskL.create(imL.size(), CV_8UC1);
		maskL.setTo(cv::Scalar(255));
	}

	auto calib = read_calibration(path/std::string("calib.txt"));
	if (imL.empty() || imR.empty() || gtL.empty() || maskL.empty()) {
		throw std::exception();
	}

	std::string name = path.filename() == "." ?
		path.parent_path().filename().c_str() :
		path.filename().c_str();

	return {name, imL, imR, gtL, maskL, calib};
}

MiddleburyData load_input(const std::string &path) {
	return load_input(fs::path(path));
}
