#include <ftl/rgbd/source.hpp>
#include "middlebury_source.hpp"
#include "cuda_algorithms.hpp"
#include <opencv2/imgcodecs.hpp>
#include <opencv2/imgproc.hpp>
#include "loader.hpp"

#include <loguru.hpp>

using ftl::rgbd::detail::MiddleburySource;
using std::string;
using ftl::codecs::Channel;
using cv::cuda::GpuMat;
using ftl::rgbd::Capability;

MiddleburySource::MiddleburySource(ftl::rgbd::Source *host)
		: ftl::rgbd::BaseSourceImpl(host), ready_(false) {
	// Not VALID
}

static bool loadMiddleburyCalib(const std::string &filename, ftl::rgbd::Camera &params, double scaling) {
	FILE* fp = fopen(filename.c_str(), "r");
	
	float cam0[3][3] = {};
	float cam1[3][3];
	float doffs = 0.0f;
	float baseline = 0.0f;
	int width = 0;
	int height = 0;
	int ndisp;
	int isint;
	int vmin;
	int vmax;
	float dyavg;
	float dymax;

	if (fp != nullptr) {
		char buff[512];
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "cam0 = [%f %f %f; %f %f %f; %f %f %f]\n", &cam0[0][0], &cam0[0][1], &cam0[0][2], &cam0[1][0], &cam0[1][1], &cam0[1][2], &cam0[2][0], &cam0[2][1], &cam0[2][2]);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "cam1 = [%f %f %f; %f %f %f; %f %f %f]\n", &cam1[0][0], &cam1[0][1], &cam1[0][2], &cam1[1][0], &cam1[1][1], &cam1[1][2], &cam1[2][0], &cam1[2][1], &cam1[2][2]);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "doffs = %f\n", &doffs);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "baseline = %f\n", &baseline);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "width = %d\n", &width);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "height = %d\n", &height);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "ndisp = %d\n", &ndisp);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "isint = %d\n", &isint);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "vmin = %d\n", &vmin);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "vmax = %d\n", &vmax);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "dyavg = %f\n", &dyavg);
		if (fgets(buff, sizeof(buff), fp) != nullptr) sscanf(buff, "dymax = %f\n", &dymax);
		fclose(fp);

		params.fx = cam0[0][0] * scaling;
		params.fy = params.fx;
		params.cx = -cam0[0][2] * scaling;
		params.cy = -cam0[1][2] * scaling;
		params.width = width * scaling;
		params.height = height * scaling;
		params.baseline = baseline;
		params.doffs = doffs * scaling;

		return true;
	}

	return false;
}

bool MiddleburySource::supported() {
	return true;
}

MiddleburySource::MiddleburySource(ftl::rgbd::Source *host, const string &dir)
		: ftl::rgbd::BaseSourceImpl(host), ready_(false) {

	double scaling = host->value("scaling", 0.5);

	if (!loadMiddleburyCalib(dir+"/calib.txt", params_, scaling)) {
		LOG(ERROR) << "Could not load middlebury calibration";
		return;
	}

    auto data = load_input(dir);

	// Add calibration to config object
	/*host_->getConfig()["focal"] = params_.fx;
	host_->getConfig()["centre_x"] = params_.cx;
	host_->getConfig()["centre_y"] = params_.cy;
	host_->getConfig()["baseline"] = params_.baseline;
	host_->getConfig()["doffs"] = params_.doffs;

	// Add event handlers to allow calibration changes...
	host_->on("baseline", [this]() {
		params_.baseline = host_->value("baseline", params_.baseline);
	});

	host_->on("focal", [this]() {
		params_.fx = host_->value("focal", params_.fx);
		params_.fy = params_.fx;
	});

	host_->on("doffs", [this]() {
		params_.doffs = host_->value("doffs", params_.doffs);
	});

	host_->on("centre_x", [this]() {
		params_.cx = host_->value("centre_x", params_.cx);
	});*/

	// left and right masks (areas outside rectified images)
	// only left mask used
	//cv::cuda::GpuMat mask_r_gpu(params_.height, params_.width, CV_8U, 255);
	//cv::cuda::GpuMat mask_l_gpu(params_.height, params_.width, CV_8U, 255);
	
	//calib_->rectifyStereo(mask_l_gpu, mask_r_gpu, stream_);
	//stream_.waitForCompletion();

	//cv::Mat mask_l;
	//mask_l_gpu.download(mask_l);
	//mask_l_ = (mask_l == 0);

	// Load image files...
	cv::Mat left_tmp, right_tmp, d_temp;
	// left_tmp = cv::imread(dir+"/im0.png", cv::IMREAD_COLOR);
	// right_tmp = cv::imread(dir+"/im1.png", cv::IMREAD_COLOR);
    cv::cvtColor(data.imL, left_tmp, cv::COLOR_BGR2BGRA);
    cv::cvtColor(data.imR, right_tmp, cv::COLOR_BGR2BGRA);

	cv::resize(left_tmp, left_tmp, cv::Size(params_.width, params_.height));
	cv::resize(right_tmp, right_tmp, cv::Size(params_.width, params_.height));
    cv::resize(data.gtL, d_temp, cv::Size(params_.width, params_.height));

	left_.upload(left_tmp, stream_);
	right_.upload(right_tmp, stream_);
    gt_.upload(d_temp, stream_);
    stream_.waitForCompletion();

    do_update_params_ = true;
}

bool MiddleburySource::capture(int64_t ts) {
	return true;
}

bool MiddleburySource::retrieve(ftl::rgbd::Frame &frame) {
    if (do_update_params_) {
		do_update_params_ = false;
		frame.setLeft() = params_;
		frame.setPose() = Eigen::Matrix4d::Identity();

		auto &meta = frame.create<std::map<std::string,std::string>>(Channel::MetaData);
		meta["name"] = host_->value("name", host_->getID());
		meta["id"] = host_->getID();
		meta["uri"] = host_->value("uri", std::string(""));
		meta["device"] = "Middlebury";

		if (!frame.has(Channel::Capabilities)) {
			auto &cap = frame.create<std::unordered_set<Capability>>(Channel::Capabilities);
		}
	}

    left_.copyTo(frame.create<GpuMat>(Channel::Colour));
    right_.copyTo(frame.create<GpuMat>(Channel::Right));
    gt_.copyTo(frame.create<GpuMat>(Channel::GroundTruth));
    return true;
}
